"""
Helps to use ujson module instead of json module.

If ujson is installed the app will use it. To prevent using installed
ujson you should use DISABLE_UJSON environment variable.
More about ujson: https://github.com/ultrajson/ultrajson

If ujson is not installed the app will use basic json module.
"""
import os

disabled = os.getenv("DISABLE_UJSON")
if not disabled:
    try:
        import ujson as json
    except ImportError:  # pragma: no cover
        import json  # type: ignore
else:  # pragma: no cover
    import json  # type: ignore


def loads(data) -> dict:
    """Load a JSON to a dict via available JSON converter."""
    return json.loads(data)


def dumps(data) -> str:  # pragma: no cover
    """Dump a data to JSON via available JSON converter."""
    return json.dumps(data, ensure_ascii=False)



"""Pydantic model represents True Checker API jobs."""

from datetime import datetime
from enum import Enum
from typing import Optional

from pydantic import BaseModel


class JobStates(str, Enum):
    """Available True Checker API job states."""

    TODO = "todo"
    IN_PROGRESS = "in_progress"
    DONE = "done"
    CANCELLED = "cancelled"
    ERROR = "error"


class CheckJob(BaseModel):  # pylint: disable=too-few-public-methods
    """True Checker API job model."""

    id: str
    bot_id: int
    state: JobStates
    progress: float
    created: datetime
    description: Optional[str]



"""Base aiohttp client class module."""

import asyncio
import io
import ssl
from pathlib import Path
from typing import Any, Dict, Optional, Union

import certifi
from aiohttp import (
    ClientSession,
    ClientTimeout,
    ContentTypeError,
    FormData,
    TCPConnector,
)
from aiohttp.typedefs import StrOrURL

from .exceptions import (
    BadRequest,
    BadState,
    TrueCheckerException,
    Unauthorized,
    ValidationError,
)
from .utils import json

EXC_MAPPING = {
    400: BadRequest,
    401: Unauthorized,
    404: BadRequest,
    409: BadState,
    422: ValidationError,
}


class BaseClient:
    """
    Base aiohttp client.

    Consists of all methods need to make a request to API:
     - session caching
     - request wrapping
     - exceptions wrapping
     - grace session close
     - e.t.c.
    """

    def __init__(self, timeout: Optional[ClientTimeout] = None):
        """
        Set defaults on object init.

        By default `self._session` is None.
        It will be created on a first API request.
        The second request will use the same `self._session`.
        """
        self._timeout = timeout or ClientTimeout(total=0)
        self._session: Optional[ClientSession] = None

    def get_session(self):
        """Get cached session. One session per instance."""
        if isinstance(self._session, ClientSession) and not self._session.closed:
            return self._session

        ssl_context = ssl.create_default_context(cafile=certifi.where())
        connector = TCPConnector(ssl=ssl_context)

        self._session = ClientSession(
            connector=connector,
            json_serialize=json.dumps,
            timeout=self._timeout,
        )
        return self._session

    async def _make_request(self, method: str, url: StrOrURL, **kwargs) -> dict:
        """
        Make a request.

        :param method: HTTP Method
        :param url: endpoint link
        :param kwargs: data, params, json and other...
        :return: status and result or exception
        """
        session = self.get_session()

        async with session.request(method, url, **kwargs) as response:
            status = response.status
            try:
                data = await response.json(loads=json.loads)
            except ContentTypeError:
                data = await response.text()

        if status != 200:
            raise self._process_exception(status, data)

        return data

    def _prepare_form(self, file: Union[str, Path, io.IOBase]) -> FormData:
        """Create form to pass file via multipart/form-data."""
        form = FormData()
        form.add_field("file", self._prepare_file(file))
        return form

    @staticmethod
    def _prepare_file(file: Union[str, Path, io.IOBase]):
        """Prepare accepted types to correct file type."""
        if isinstance(file, str):
            return Path(file).open("rb")

        if isinstance(file, io.IOBase):
            return file

        if isinstance(file, Path):
            return file.open("rb")

        raise TypeError(f"Not supported file type: `{type(file).__name__}`")

    @staticmethod
    def _process_exception(
        status: int, data: Union[Dict[str, Any], str]
    ) -> TrueCheckerException:
        """
        Wrap API exceptions.

        :param status: response status
        :param data: response json converted to dict()
        :return: wrapped exception
        """
        if isinstance(data, dict):
            text = data.get("message") or data.get("detail")
        else:
            text = data
        exc = EXC_MAPPING.get(status, TrueCheckerException)
        return exc(text)

    async def close(self):
        """Close the session graceful."""
        if not isinstance(self._session, ClientSession):
            return

        if self._session.closed:
            return

        await self._session.close()
        await asyncio.sleep(0.25)



"""Pydantic model represents Telegram bot profile."""

from datetime import datetime

from pydantic import BaseModel

from .users import Users


class Profile(BaseModel):  # pylint: disable=too-few-public-methods
    """Telegram bot profile model."""

    username: str
    full_name: str
    updated: datetime
    users: Users



"""True Checker API exceptions."""


class TrueCheckerException(Exception):
    """Base (root) True Checker exception."""


class BadRequest(TrueCheckerException):
    """
    Base Bad Request exception.

    Status: 400
    """


class Unauthorized(BadRequest):
    """
    Unauthorized exception.

    Occurs on troubles with Telegram token.
    Status: 401
    """


class BadState(BadRequest):
    """
    State conflict exception.

    Occurs if current state is not fits for request.
    Status: 409
    """


class ValidationError(BadRequest):
    """
    Wrong params are passed to API method.

    Status: 422
    """


class NotFound(BadRequest):
    """
    Requested object is not found.

    Status: 404
    """



"""
Python client library for True Checker API.

Repository: https://gitlab.com/true-web-app/true-checker/true-checker-python
API docs: https://checker.trueweb.app/redoc
"""

__all__ = ["TrueChecker"]

import os

from .api import TrueChecker

# install uvloop if exists and not disabled
try:
    import uvloop  # noqa
except ImportError:  # pragma: no cover
    pass
else:
    if "DISABLE_UVLOOP" not in os.environ:
        uvloop.install()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['truechecker', 'truechecker.models', 'truechecker.utils']

package_data = \
{'': ['*']}

install_requires = \
['aiohttp>=3.7,<4.0', 'certifi>=2021,<2022', 'pydantic>=1.8,<2.0']

setup_kwargs = {
    'name': 'truechecker',
    'version': '2021.10.9',
    'description': 'Python client for True Checker API',
    'long_description': '# True Checker for Python\n\nPython client library for [True Checker API](https://checker.trueweb.app/redoc)\n\n## How to install\n\n```bash\npip install truechecker\n```\n\n## How to use\n\n```python\n# import TrueChecker\nfrom truechecker import TrueChecker\n\n\n# create an instance (poss your Telegram bot token here)\nchecker = TrueChecker("your_bot_token")\n\n\n# prepare a file with users ids\n# you should use string path, pathlib.Path object or io.BaseFile\nfile_path = "downloads/users.csv"\n\n\n# send request to create a new job\njob = await checker.check_profile(file_path)\nprint("Job created. ID:", job.id)\n\n\n# get the status of job\njob = await checker.get_job_status(job.id)\nprint("Job state:", job.state)\nprint("Job progress:", job.progress)\n\n\n# if the job is done, let\'s get the profile\nprofile = await checker.get_profile("my_bot_username")\nprint("Bot profile:", profile)\n\n\n# if you need to cancel the job\njob = await checker.cancel_job(job.id)\nprint("Job state:", job.state)  # Cancelled\n\n\n# Don\'t forget to close checker on your app\'s on_shutdown\nawait checker.close()\n\n```\n_CAUTION: it\'s not a full code example. Await statements should be used within coroutines only._\n\n## Contributing\nBefore making Pull/Merge Requests, please read the [Contributing guidelines](CONTRIBUTING.md)\n',
    'author': 'Oleg A.',
    'author_email': 'oleg@trueweb.app',
    'maintainer': 'Oleg A.',
    'maintainer_email': 'oleg@trueweb.app',
    'url': 'https://gitlab.com/true-web-app/true-checker/true-checker-python',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



"""Pydantic model represents Telegram bot users statistics."""

from pydantic import BaseModel


class Users(BaseModel):  # pylint: disable=too-few-public-methods
    """Telegram bot users statistics model (part of Profile)."""

    active: int
    stopped: int
    deleted: int
    not_found: int



"""True Checker API client module."""

import io
from pathlib import Path
from typing import Optional, Union

from aiohttp import ClientTimeout

from .base import BaseClient
from .const import HTTPMethods
from .models import CheckJob, Profile


class TrueChecker(BaseClient):
    """
    True Checker API client.

    Consists of API methods only.
    All other methods are hidden in the BaseClient.
    """

    API_VERSION = "1.1.0"
    API_HOST = "https://checker.trueweb.app/api"
    API_DOCS = "https://checker.trueweb.app/redoc"

    def __init__(
        self,
        token: str,
        api_host: Optional[str] = None,
        timeout: Optional[ClientTimeout] = None,
    ):
        """
        Init True Checker API client.

        :param token: Telegram bot token
        :param api_host: API host url (optional)
        :param timeout: set session timeout, default is None
        """
        super().__init__(timeout=timeout)
        self.__token = token
        self._api_host = api_host or self.API_HOST

    async def check_profile(
        self,
        file: Union[str, Path, io.IOBase],
        delay: Optional[float] = None,
    ) -> CheckJob:
        """
        Init new bot checking.

        :param file: Any format of user_id array (csv, one-per-line or
            other). Telegram IDs will be parsed via regexp, so don't
            provide other digits looks like an telegram chat ID.
        :param delay: Delay between typing requests
        :return: new job object
        """
        method = HTTPMethods.PUT
        url = f"{self._api_host}/profile/{self.__token}"
        form = self._prepare_form(file)

        params = {}
        if delay is not None:
            params["delay"] = delay

        data = await self._make_request(method, url, params=params, data=form)
        return CheckJob(**data)

    async def get_profile(self, username: str) -> Profile:
        """
        Return checked bot profile on success.

        :param username: Bot username. Case insensitive
        :return: bot profile object or not found exception
        """
        method = HTTPMethods.GET
        url = f"{self._api_host}/profile/{username}"

        data = await self._make_request(method, url)
        return Profile(**data)

    async def get_job_status(self, job_id: str) -> CheckJob:
        """
        Find job by id and return job status.

        :param job_id: Job ID received after `check_profile` method
        :return: job object or not found exception
        """
        method = HTTPMethods.GET
        url = f"{self._api_host}/job/{job_id}"

        data = await self._make_request(method, url)
        return CheckJob(**data)

    async def cancel_job(self, job_id: str) -> CheckJob:
        """
        Cancel running Job.

        Job will stop with minor delay.
        (it may continue working for some additional seconds)

        :param job_id: Job ID received after `check_profile` method
        :return: job object / not found / not running
        """
        method = HTTPMethods.DELETE
        url = f"{self._api_host}/job/{job_id}"

        data = await self._make_request(method, url)
        return CheckJob(**data)
